import numpy as np
import pdb2sql
import os
import pickle


def _printif(string, cond): return print(string) if cond else None


class StructureSimilarity(object):

    def __init__(self, decoy, ref, verbose=False):
        """Compute the structure similarity between different molecules.

        This class allows to compute the i-RMSD, L-RMSD, Fnat and DockQ score of a given conformation.
        This can be a replacement for ProFIT. Note that the calculation of the zones are done by the class itself
        and does not require any extra input. This local class could be replaced by pdb2sql.StructureSimilarity.
        (https://github.com/DeepRank/pdb2sql)

        Args:
            decoy (str): file name of the decoy
            ref (str): file name of the reference
            verbose (bool, optional): print debug information


        Example :

        >>> from deeprank.tools import StructureSimilarity
        >>> decoy = '1AK4_1w.pdb'
        >>> ref = '1AK4.pdb'
        >>> sim = StructureSimilarity(decoy,ref)
        >>> irmsd_fast = sim.compute_irmsd_fast(method='svd',izone='1AK4.izone')
        >>> irmsd = sim.compute_irmsd_pdb2sql(method='svd',izone='1AK4.izone')
        >>> lrmsd_fast = sim.compute_lrmsd_fast(method='svd',lzone='1AK4.lzone',check=True)
        >>> lrmsd = sim.compute_lrmsd_pdb2sql(exportpath=None,method='svd')
        >>> Fnat = sim.compute_Fnat_pdb2sql()
        >>> Fnat_fast = sim.compute_Fnat_fast(ref_pairs='1AK4.ref_pairs')
        >>> dockQ = sim.compute_DockQScore(Fnat_fast,lrmsd_fast,irmsd_fast)

        """

        self.decoy = decoy
        self.ref = ref
        self.verbose = verbose

    def compute_lrmsd_fast(self, lzone=None, method='svd', check=True):
        """Fast routine to compute the L-RMSD.

        This routine parse the PDB directly without using pdb2sql.
        L-RMSD is computed by aligning the longest chain of the decoy to the one of the reference
        and computing the RMSD of the shortest chain between decoy and reference.

        Ref : DockQ: A Quality Measure for Protein-Protein Docking Models
              http://journals.plos.org/plosone/article?id=10.1371/journal.pone.0161879

        Args:
            lzone (None, optional): name of the file containing the zone definition. If None the file will be calculated first
            method (str, optional): Method to align the fragments ('svd','quaternion')
            check (bool, optional): Check if the sequences are aligned and fix it if not. Should be True

        Returns:
            float: L-RMSD value of the conformation
        """

        # create/read the lzone file
        if lzone is None:
            resData = self.compute_lzone(save_file=False)
        elif not os.path.isfile(lzone):
            self.compute_lzone(save_file=True, filename=lzone)
            resData = self.read_zone(lzone)
        else:
            resData = self.read_zone(lzone)

        ##################################################
        # the check make sure that all the
        # atoms are in the correct order
        # I STRONGLY discourage turning the check off
        # it actually reorder the xyz data of the native/decoy
        # to match.
        ##################################################
        if check:

            data_decoy_long, data_decoy_short = self.read_data_zone(
                self.decoy, resData, return_not_in_zone=True)
            data_ref_long,   data_ref_short = self.read_data_zone(
                self.ref, resData, return_not_in_zone=True)

            atom_decoy_long = [data[:3] for data in data_decoy_long]
            atom_ref_long = [data[:3] for data in data_ref_long]

            xyz_decoy_long, xyz_ref_long = [], []
            for ind_decoy, at in enumerate(atom_decoy_long):

                try:
                    ind_ref = atom_ref_long.index(at)
                    xyz_decoy_long.append(
                        data_decoy_long[ind_decoy][3:])
                    xyz_ref_long.append(data_ref_long[ind_ref][3:])
                except ValueError:
                    pass

            atom_decoy_short = [data[:3] for data in data_decoy_short]
            atom_ref_short = [data[:3] for data in data_ref_short]

            xyz_decoy_short, xyz_ref_short = [], []
            for ind_decoy, at in enumerate(atom_decoy_short):
                try:
                    ind_ref = atom_ref_short.index(at)
                    xyz_decoy_short.append(
                        data_decoy_short[ind_decoy][3:])
                    xyz_ref_short.append(data_ref_short[ind_ref][3:])
                except ValueError:
                    pass

        # extract the xyz
        else:
            print('WARNING : The atom order have not been checked. Switch to check=True or continue at your own risk')
            xyz_decoy_long, xyz_decoy_short = self.read_xyz_zone(
                self.decoy, resData, return_not_in_zone=True)
            xyz_ref_long, xyz_ref_short = self.read_xyz_zone(
                self.ref, resData, return_not_in_zone=True)

        # get the translation so that both A chains are centered
        tr_decoy = self.get_trans_vect(xyz_decoy_long)
        tr_ref = self.get_trans_vect(xyz_ref_long)

        # translate everything for 1
        xyz_decoy_short = self.translation(xyz_decoy_short, tr_decoy)
        xyz_decoy_long = self.translation(xyz_decoy_long, tr_decoy)

        # translate everuthing for 2
        xyz_ref_short = self.translation(xyz_ref_short, tr_ref)
        xyz_ref_long = self.translation(xyz_ref_long, tr_ref)

        # get the ideql rotation matrix
        # to superimpose the A chains
        U = self.get_rotation_matrix(
            xyz_decoy_long, xyz_ref_long, method=method)

        # rotate the entire fragment
        xyz_decoy_short = self.rotation_matrix(
            xyz_decoy_short, U, center=False)

        # compute the RMSD
        return self.get_rmsd(xyz_decoy_short, xyz_ref_short)

    def compute_lzone(self, save_file=True, filename=None):
        """Compute the zone for L-RMSD calculation

        Args:
            save_file (bool, optional): save the zone file
            filename (str, optional): name of the file

        Returns:
            dict: definition of the zone
        """
        sql_ref = pdb2sql(self.ref)
        nA = len(sql_ref.get('x,y,z', chainID='A'))
        nB = len(sql_ref.get('x,y,z', chainID='B'))

        # detect which chain is the longest
        long_chain = 'A'
        if nA < nB:
            long_chain = 'B'

        # extract data about the residue
        data_test = [tuple(data) for data in sql_ref.get(
            'chainID,resSeq', chainID=long_chain)]
        data_test = sorted(set(data_test))

        # close the sql
        sql_ref.close()

        if save_file:
            if filename is None:
                f = open(self.ref.split('.')[0]+'.lzone', 'w')
            else:
                f = open(filename, 'w')
            for res in data_test:
                chain = res[0]
                num = res[1]
                f.write('zone %s%d-%s%d\n' % (chain, num, chain, num))
            f.close()
            return

        else:
            resData = {}
            for res in data_test:
                chain = res[0]
                num = res[1]

                if chain not in resData.keys():
                    resData[chain] = []
                resData[chain].append(num)
            return resData

    def compute_irmsd_fast(self, izone=None, method='svd', cutoff=10.0, check=True):
        """Fast method to compute the i-rmsd

        Require the precalculation of the izone
        A dedicated routine is implemented to comoute the izone
        if izone is not given in argument the routine will compute them automatically

        i-RMSD is computed selecting the back bone of the contact residue with a cutoff of 10A
        in the decoy. Align these as best as possible with their coutner part in the ref
        and and compute the RMSD

        Ref : DockQ: A Quality Measure for Protein-Protein Docking Models
              http://journals.plos.org/plosone/article?id=10.1371/journal.pone.0161879

        Args:
            izone (None, optional): file name of the zone. if None the zones will be calculated first
            method (str, optional): Method to align the fragments ('svd','quaternion')
            cutoff (float, optional): cutoff for the contact atoms
            check (bool, optional): Check if the sequences are aligned and fix it if not. Should be True

        Returns:
            float: i-RMSD value of the conformation
        """

        # read the izone file
        if izone is None:
            resData = self.compute_izone(cutoff, save_file=False)
        elif not os.path.isfile(izone):
            self.compute_izone(cutoff, save_file=True, filename=izone)
            resData = self.read_zone(izone)
        else:
            resData = self.read_zone(izone)

        ##################################################
        # the check make sure that all the
        # atoms are in the correct order
        # I STRONGLY discourage turning the check off
        # it actually fixes the order
        ##################################################
        if check:

            data_decoy = self.read_data_zone(
                self.decoy, resData, return_not_in_zone=False)
            data_ref = self.read_data_zone(
                self.ref, resData, return_not_in_zone=False)

            atom_decoy = [data[:3] for data in data_decoy]
            atom_ref = [data[:3] for data in data_ref]

            xyz_contact_decoy, xyz_contact_ref = [], []
            for ind_decoy, at in enumerate(atom_decoy):
                try:
                    ind_ref = atom_ref.index(at)
                    xyz_contact_decoy.append(
                        data_decoy[ind_decoy][3:])
                    xyz_contact_ref.append(data_ref[ind_ref][3:])
                except ValueError:
                    pass

        # extract the xyz
        else:
            print('WARNING : The atom order have not been checked. Switch to check=True or continue at your own risk')
            xyz_contact_decoy = self.read_xyz_zone(
                self.decoy, resData)
            xyz_contact_ref = self.read_xyz_zone(self.ref, resData)

        # get the translation so that both A chains are centered
        tr_decoy = self.get_trans_vect(xyz_contact_decoy)
        tr_ref = self.get_trans_vect(xyz_contact_ref)

        # translate everything
        xyz_contact_decoy = self.translation(
            xyz_contact_decoy, tr_decoy)
        xyz_contact_ref = self.translation(xyz_contact_ref, tr_ref)

        # get the ideql rotation matrix
        # to superimpose the A chains
        U = self.get_rotation_matrix(
            xyz_contact_decoy, xyz_contact_ref, method=method)

        # rotate the entire fragment
        xyz_contact_decoy = self.rotation_matrix(
            xyz_contact_decoy, U, center=False)

        # return the RMSD
        return self.get_rmsd(xyz_contact_decoy, xyz_contact_ref)

    def compute_izone(self, cutoff=5.0, save_file=True, filename=None):
        """Compute the zones  for i-rmsd calculationss

        Args:
            cutoff (float, optional): cutoff for the contact atoms
            save_file (bool, optional): svae file containing the zone
            filename (str, optional): filename

        Returns:
            dict: i-zone definition
        """

        sql_ref = pdb2sql(self.ref)
        contact_ref = sql_ref.get_contact_atoms(
            cutoff=cutoff, extend_to_residue=True, return_only_backbone_atoms=True)
        index_contact_ref = contact_ref[0]+contact_ref[1]

        # get the xyz and atom identifier of the decoy contact atoms
        #xyz_contact_ref = sql_ref.get('x,y,z',rowID=index_contact_ref)
        data_test = [tuple(data) for data in sql_ref.get(
            'chainID,resSeq', rowID=index_contact_ref)]
        data_test = sorted(set(data_test))

        # close the sql
        sql_ref.close()

        if save_file:

            if filename is None:
                f = open(self.ref.split('.')[0]+'.izone', 'w')
            else:
                f = open(filename, 'w')

            for res in data_test:
                chain = res[0]
                num = res[1]
                f.write('zone %s%d-%s%d\n' % (chain, num, chain, num))
            f.close()
            return

        else:
            resData = {}
            for res in data_test:
                chain = res[0]
                num = res[1]

                if chain not in resData.keys():
                    resData[chain] = []
                resData[chain].append(num)
            return resData

    def compute_Fnat_fast(self, ref_pairs=None, cutoff=5):
        """Compute the FNAT of the conformation

        Args:
            ref_pairs (str, optional): file name describing the pairs
            cutoff (int, optional): cutoff for the contact atoms

        Returns:
            float: FNAT value

        Raises:
            ValueError: if the decoy file is not found
        """
        # read the izone file
        if ref_pairs is None:
            residue_pairs_ref = self.compute_residue_pairs_ref(
                cutoff, save_file=False)

        elif not os.path.isfile(ref_pairs):
            self.compute_residue_pairs_ref(
                cutoff, save_file=True, filename=ref_pairs)
            f = open(ref_pairs, 'rb')
            residue_pairs_ref = pickle.load(f)
            f.close()
        else:
            f = open(ref_pairs, 'rb')
            residue_pairs_ref = pickle.load(f)
            f.close()

        # create a dict of the ecoy data
        if isinstance(self.decoy, str) and os.path.isfile(self.decoy):
            with open(self.decoy, 'r') as f:
                data_decoy = f.readlines()
            decoy_name = os.path.basename(self.decoy)
        elif isinstance(self.decoy, np.ndarray):
            data_decoy = [l.decode('utf-8') for l in self.decoy]
            decoy_name = 'decoy'
        else:
            raise ValueError('Decoy not found')

        # read the decoy data
        atom_decoy, xyz_decoy = [], []
        residue_xyz = {}
        residue_name = {}

        # go through all the lines
        # that starts with ATOM
        for line in data_decoy:

            if line.startswith('ATOM'):

                # chain ID
                chainID = line[21]
                if chainID == ' ':
                    chainID = line[72]

                # atom info
                resSeq = int(line[22:26])
                resName = line[17:20].strip()
                name = line[12:16].strip()

                # position
                x, y, z = float(line[30:38]), float(
                    line[38:46]), float(line[46:54])

                # dict entry
                key = (chainID, resSeq, resName)

                # create the dict entry if necessary
                if key not in residue_xyz.keys():
                    residue_xyz[key] = []
                    residue_name[key] = []

                # we exclude the Hydrogens from the search
                if name[0] != 'H':
                    residue_xyz[key].append([x, y, z])
                    residue_name[key].append(name)

        # loop over the residue pairs of the
        # and increment common if an atom pair is close enough
        nCommon, nTotal = 0, 0
        for resA, resB_list in residue_pairs_ref.items():
            if resA in residue_xyz:
                xyzA = residue_xyz[resA]
                for resB in resB_list:
                    if resB in residue_xyz.keys():
                        xyzB = residue_xyz[resB]
                        dist_min = np.min(np.array([np.sqrt(np.sum(
                            (np.array(p1)-np.array(p2))**2)) for p1 in xyzA for p2 in xyzB]))
                        if dist_min <= cutoff:
                            nCommon += 1
                    nTotal += 1
            else:
                msg = '\t FNAT Warning could not find residue: ', resA, ' in: ', decoy_name
                _printif(msg, self.verbose)

        # normalize
        return nCommon/nTotal

    def compute_residue_pairs_ref(self, cutoff=5.0, save_file=True, filename=None):
        """Compute the residue pair on the reference conformation

        Args:
            cutoff (float, optional): cutoff for the contact atoms
            save_file (bool, optional): save the file containing the residue pairs
            filename (None, optional): filename

        Returns:
            dict: defintition of the residue pairs
        """
        sql_ref = pdb2sql(self.ref, sqlfile='mol2.db')
        residue_pairs_ref = sql_ref.get_contact_residue(cutoff=cutoff, return_contact_pairs=True,
                                                        excludeH=True)
        sql_ref.close()

        if save_file:
            if filename is None:
                f = open(self.ref.split('.')[
                         0]+'residue_contact_pairs.pckl', 'wb')
            else:
                f = open(filename, 'wb')

            # save as pickle
            pickle.dump(residue_pairs_ref, f)
            f.close()
            return
        else:
            return residue_pairs_ref

    def compute_lrmsd_pdb2sql(self, exportpath=None, method='svd'):
        """Slow routine to compute the L-RMSD.

        This routine parse the PDB directly using pdb2sql.
        L-RMSD is computed by aligning the longest chain of the decoy to the one of the reference
        and computing the RMSD of the shortest chain between decoy and reference.

        Ref : DockQ: A Quality Measure for Protein-Protein Docking Models
              http://journals.plos.org/plosone/article?id=10.1371/journal.pone.0161879

        Args:
            exportpath (str, optional): file name where the aligned pdbs are exported
            method (str, optional): Method to align the fragments ('svd','quaternion')

        Returns:
            float: L-RMSD value of the conformation
        """

        # create the sql
        sql_decoy = pdb2sql(self.decoy, sqlfile='decoy.db')
        sql_ref = pdb2sql(self.ref, sqlfile='ref.db')

        # extract the pos of chains A
        xyz_decoy_A = np.array(sql_decoy.get('x,y,z', chainID='A'))
        xyz_ref_A = np.array(sql_ref.get('x,y,z', chainID='A'))

        # extract the pos of chains B
        xyz_decoy_B = np.array(sql_decoy.get('x,y,z', chainID='B'))
        xyz_ref_B = np.array(sql_ref.get('x,y,z', chainID='B'))

        # check the lengthes
        if len(xyz_decoy_A) != len(xyz_ref_A):
            xyz_decoy_A, xyz_ref_A = self.get_identical_atoms(
                sql_decoy, sql_ref, 'A')

        if len(xyz_decoy_B) != len(xyz_ref_B):
            xyz_decoy_B, xyz_ref_B = self.get_identical_atoms(
                sql_decoy, sql_ref, 'B')

        # detect which chain is the longest
        nA, nB = len(xyz_decoy_A), len(xyz_decoy_B)
        if nA > nB:
            xyz_decoy_long = xyz_decoy_A
            xyz_ref_long = xyz_ref_A

            xyz_decoy_short = xyz_decoy_B
            xyz_ref_short = xyz_ref_B

        else:
            xyz_decoy_long = xyz_decoy_B
            xyz_ref_long = xyz_ref_B

            xyz_decoy_short = xyz_decoy_A
            xyz_ref_short = xyz_ref_A

        # get the translation so that both A chains are centered
        tr_decoy = self.get_trans_vect(xyz_decoy_long)
        tr_ref = self.get_trans_vect(xyz_ref_long)

        # translate everything for 1
        xyz_decoy_short = self.translation(xyz_decoy_short, tr_decoy)
        xyz_decoy_long = self.translation(xyz_decoy_long, tr_decoy)

        # translate everuthing for 2
        xyz_ref_short = self.translation(xyz_ref_short, tr_ref)
        xyz_ref_long = self.translation(xyz_ref_long, tr_ref)

        # get the ideal rotation matrix
        # to superimpose the A chains
        U = self.get_rotation_matrix(
            xyz_decoy_long, xyz_ref_long, method=method)

        # rotate the entire fragment
        xyz_decoy_short = self.rotation_matrix(
            xyz_decoy_short, U, center=False)

        # compute the RMSD
        lrmsd = self.get_rmsd(xyz_decoy_short, xyz_ref_short)

        # export the pdb for verifiactions
        if exportpath is not None:

            # extract the pos of the dimer
            xyz_decoy = np.array(sql_decoy.get('x,y,z'))
            xyz_ref = np.array(sql_ref.get('x,y,z'))

            # translate
            xyz_ref = self.translation(xyz_ref, tr_ref)
            xyz_decoy = self.translation(xyz_decoy, tr_decoy)

            # rotate decoy
            xyz_decoy = self.rotation_matrix(
                xyz_decoy, U, center=False)

            # update the sql database
            sql_decoy.update_xyz(xyz_decoy)
            sql_ref.update_xyz(xyz_ref)

            # export
            sql_decoy.exportpdb(exportpath+'/lrmsd_decoy.pdb')
            sql_ref.exportpdb(exportpath+'/lrmsd_aligned.pdb')

        # close the db
        sql_decoy.close()
        sql_ref.close()

        return lrmsd

    @staticmethod
    def get_identical_atoms(db1, db2, chain):
        """Return that atoms shared by both databse for a specific chain

        Args:
            db1 (TYPE): pdb2sql database of the first conformation
            db2 (TYPE): pdb2sql database of the 2nd conformation
            chain (str): chain name

        Returns:
            list, list: list of xyz for both database
        """
        # get data
        data1 = db1.get('chainID,resSeq,name', chainID=chain)
        data2 = db2.get('chainID,resSeq,name', chainID=chain)

        # tuplify
        data1 = [tuple(d1) for d1 in data1]
        data2 = [tuple(d2) for d2 in data2]

        # get the intersection
        shared_data = list(set(data1).intersection(data2))

        # get the xyz
        xyz1, xyz2 = [], []
        for data in shared_data:
            query = 'SELECT x,y,z from ATOM WHERE chainID=? AND resSeq=? and name=?'
            xyz1.append(list(list(db1.c.execute(query, data))[0]))
            xyz2.append(list(list(db2.c.execute(query, data))[0]))

        return xyz1, xyz2

    def compute_irmsd_pdb2sql(self, cutoff=10, method='svd', izone=None, exportpath=None):
        """Slow method to compute the i-rmsd

        Require the precalculation of the izone. A dedicated routine is implemented to comoute the izone
        if izone is not given in argument the routine will compute them automatically

        i-RMSD is computed selecting the back bone of the contact residue with a cutoff of 10A
        in the decoy. Align these as best as possible with their coutner part in the ref
        and and compute the RMSD

        Ref : DockQ: A Quality Measure for Protein-Protein Docking Models
              http://journals.plos.org/plosone/article?id=10.1371/journal.pone.0161879

        Args:
            izone (None, optional): file name of the zone. if None the zones will be calculated first
            method (str, optional): Method to align the fragments ('svd','quaternion')
            cutoff (float, optional): cutoff for the contact atoms
            exportpath (str, optional): file name where the aligned pdbs are exported

        Returns:
            float: i-RMSD value of the conformation
        """

        # create thes sql
        sql_decoy = pdb2sql(self.decoy)
        sql_ref = pdb2sql(self.ref)

        # get the contact atoms
        if izone is None:
            contact_ref = sql_ref.get_contact_atoms(cutoff=cutoff, extend_to_residue=True,
                                                    return_only_backbone_atoms=False)
            index_contact_ref = contact_ref[0]+contact_ref[1]
        else:
            index_contact_ref = self.get_izone_rowID(
                sql_ref, izone, return_only_backbone_atoms=False)

        # get the xyz and atom identifier of the decoy contact atoms
        xyz_contact_ref = sql_ref.get(
            'x,y,z', rowID=index_contact_ref)
        data_contact_ref = sql_ref.get(
            'chainID,resSeq,resName,name', rowID=index_contact_ref)

        # get the xyz and atom indeitifier of the reference
        xyz_decoy = sql_decoy.get('x,y,z')
        data_decoy = sql_decoy.get('chainID,resSeq,resName,name')

        # loop through the ref label
        # check if the atom is in the decoy
        # if yes -> add xyz to xyz_contact_decoy
        # if no  -> remove the corresponding to xyz_contact_ref
        xyz_contact_decoy = []
        index_contact_decoy = []
        clean_ref = False
        for iat, atom in enumerate(data_contact_ref):

            try:
                index = data_decoy.index(atom)
                index_contact_decoy.append(index)
                xyz_contact_decoy.append(xyz_decoy[index])
            except Exception:
                xyz_contact_ref[iat] = None
                index_contact_ref[iat] = None
                clean_ref = True

        # clean the xyz
        if clean_ref:
            xyz_contact_ref = [
                xyz for xyz in xyz_contact_ref if xyz is not None]
            index_contact_ref = [
                ind for ind in index_contact_ref if ind is not None]

        # check that we still have atoms in both chains
        chain_decoy = list(
            set(sql_decoy.get('chainID', rowID=index_contact_decoy)))
        chain_ref = list(
            set(sql_ref.get('chainID', rowID=index_contact_ref)))

        if len(chain_decoy) < 1 or len(chain_ref) < 1:
            raise ValueError(
                'Error in i-rmsd: only one chain represented in one chain')

        # get the translation so that both A chains are centered
        tr_decoy = self.get_trans_vect(xyz_contact_decoy)
        tr_ref = self.get_trans_vect(xyz_contact_ref)

        # translate everything
        xyz_contact_decoy = self.translation(
            xyz_contact_decoy, tr_decoy)
        xyz_contact_ref = self.translation(xyz_contact_ref, tr_ref)

        # get the ideql rotation matrix
        # to superimpose the A chains
        U = self.get_rotation_matrix(
            xyz_contact_decoy, xyz_contact_ref, method=method)

        # rotate the entire fragment
        xyz_contact_decoy = self.rotation_matrix(
            xyz_contact_decoy, U, center=False)

        # compute the RMSD
        irmsd = self.get_rmsd(xyz_contact_decoy, xyz_contact_ref)


        # export the pdb for verifiactions
        if exportpath is not None:

            # update the sql database
            sql_decoy.update_xyz(
                xyz_contact_decoy, index=index_contact_decoy)
            sql_ref.update_xyz(
                xyz_contact_ref, index=index_contact_ref)

            sql_decoy.exportpdb(
                exportpath+'/irmsd_decoy.pdb', index=index_contact_decoy)
            sql_ref.exportpdb(
                exportpath+'/irmsd_ref.pdb', index=index_contact_ref)

        # close the db
        sql_decoy.close()
        sql_ref.close()

        return irmsd

    @staticmethod
    def get_izone_rowID(sql, izone, return_only_backbone_atoms=True):
        """Compute the index of the izone atoms

        Args:
            sql (pdb2sql): database of the conformation
            izone (str): filename to store the zone
            return_only_backbone_atoms (bool, optional): Returns only the backbone atoms

        Returns:
            lis(int): index of the atoms in the zone

        Raises:
            FileNotFoundError: if the izone file is not found
        """
        # read the file
        if not os.path.isfile(izone):
            raise FileNotFoundError('i-zone file not found', izone)

        with open(izone, 'r') as f:
            data = f.readlines()

        # get the data out of it
        resData = {}
        for line in data:

            res = line.split()[1].split('-')[0]
            chainID, resSeq = res[0], int(res[1:])

            if chainID not in resData.keys():
                resData[chainID] = []

            resData[chainID].append(resSeq)

        # get the rowID
        index_contact = []

        for chainID, resSeq in resData.items():
            if return_only_backbone_atoms:
                index_contact += sql.get('rowID', chainID=chainID, resSeq=resSeq,
                                         name=['C', 'CA', 'N', 'O'])
            else:
                index_contact += sql.get('rowID',
                                         chainID=chainID, resSeq=resSeq)

        return index_contact

    def compute_Fnat_pdb2sql(self, cutoff=5.0):
        """Slow method to compute the FNAT usign pdb2sql

        Args:
            cutoff (float, optional): cutoff for the contact atoms

        Returns:
            float: Fnat value for the conformation
        """
        # create the sql
        sql_decoy = pdb2sql(self.decoy)
        sql_ref = pdb2sql(self.ref)

        # get the contact atoms of the decoy
        residue_pairs_decoy = sql_decoy.get_contact_residue(cutoff=cutoff,
                                                            return_contact_pairs=True,
                                                            excludeH=True)

        # get the contact atoms of the ref
        residue_pairs_ref = sql_ref.get_contact_residue(cutoff=cutoff,
                                                        return_contact_pairs=True,
                                                        excludeH=True)

        # form the pair data
        data_pair_decoy = []
        for resA, resB_list in residue_pairs_decoy.items():
            data_pair_decoy += [(resA, resB) for resB in resB_list]

        # form the pair data
        data_pair_ref = []
        for resA, resB_list in residue_pairs_ref.items():
            data_pair_ref += [(resA, resB) for resB in resB_list]

        # find the umber of residue that ref and decoys hace in common
        nCommon = len(
            set(data_pair_ref).intersection(data_pair_decoy))

        # normalize
        Fnat = nCommon/len(data_pair_ref)

        sql_decoy.close()
        sql_ref.close()

        return Fnat


    ################################################################################################
    #
    #   HELPER ROUTINES TO HANDLE THE ZONE FILES
    #
    #################################################################################################

    @staticmethod
    def read_xyz_zone(pdb_file, resData, return_not_in_zone=False):
        """Read the xyz of the zone atoms

        Args:
            pdb_file (str): filename containing the pdb of the molecule
            resData (dict): information about the residues
            return_not_in_zone (bool, optional): Do we return the atoms not in the zone

        Returns:
            list(float): XYZ of the atoms in the zone
        """
        # read the ref file
        with open(pdb_file, 'r') as f:
            data = f.readlines()

        # get the xyz of the
        xyz_in_zone = []
        xyz_not_in_zone = []

        for line in data:

            if line.startswith('ATOM'):

                chainID = line[21]
                if chainID == ' ':
                    chainID = line[72]

                resSeq = int(line[22:26])
                name = line[12:16].strip()

                x = float(line[30:38])
                y = float(line[38:46])
                z = float(line[46:54])

                if chainID in resData.keys():

                    if resSeq in resData[chainID] and name in ['C', 'CA', 'N', 'O']:
                        xyz_in_zone.append([x, y, z])

                    elif resSeq not in resData[chainID] and name in ['C', 'CA', 'N', 'O']:
                        xyz_not_in_zone.append([x, y, z])

                else:
                    if name in ['C', 'CA', 'N', 'O']:
                        xyz_not_in_zone.append([x, y, z])

        if return_not_in_zone:
            return xyz_in_zone, xyz_not_in_zone

        else:
            return xyz_in_zone

    @staticmethod
    def read_data_zone(pdb_file, resData, return_not_in_zone=False):
        """Read the data of the atoms in the zone.

        Args:
            pdb_file (str): filename containing the pdb of the molecule
            resData (dict): information about the residues
            return_not_in_zone (bool, optional): Do we return the atoms not in the zone

        Returns:
            list(float): data of the atoms in the zone
        """
        # read the ref file
        if isinstance(pdb_file, str) and os.path.isfile(pdb_file):
            with open(pdb_file, 'r') as f:
                data = f.readlines()
        elif isinstance(pdb_file, np.ndarray):
            data = [l.decode('utf-8') for l in pdb_file]

        # get the xyz of the
        data_in_zone = []
        data_not_in_zone = []

        for line in data:

            if line.startswith('ATOM'):

                chainID = line[21]
                if chainID == ' ':
                    chainID = line[72]

                resSeq = int(line[22:26])
                name = line[12:16].strip()

                x = float(line[30:38])
                y = float(line[38:46])
                z = float(line[46:54])

                if chainID in resData.keys():

                    if resSeq in resData[chainID] and name in ['C', 'CA', 'N', 'O']:
                        data_in_zone.append(
                            [chainID, resSeq, name, x, y, z])

                    elif resSeq not in resData[chainID] and name in ['C', 'CA', 'N', 'O']:
                        data_not_in_zone.append(
                            [chainID, resSeq, name, x, y, z])

                else:
                    if name in ['C', 'CA', 'N', 'O']:
                        data_not_in_zone.append(
                            [chainID, resSeq, name, x, y, z])

        if return_not_in_zone:
            return data_in_zone, data_not_in_zone

        else:
            return data_in_zone

    @staticmethod
    def read_zone(zone_file):
        """Read the zone file.

        Args:
            zone_file (str): name of the file

        Returns:
            dict: Info aboyt the residues in the zone

        Raises:
            FileNotFoundError: if the zone file is not found
        """
        # read the izone file
        if not os.path.isfile(zone_file):
            raise FileNotFoundError('zone file not found', zone_file)

        with open(zone_file, 'r') as f:
            data = f.readlines()

        # get the data out of it
        resData = {}
        for line in data:

            #    line = zone A4-A4   for positive resNum
            # or line = zone A-4-A-4 for negative resNum
            # that happens for example in 2OUL

            # split the line
            res = line.split()[1].split('-')

            # if the resnum was positive
            # we have e.g res = [A4,A4]
            if len(res) == 2:
                res = res[0]
                chainID, resSeq = res[0], int(res[1:])

            # if the resnum was negative was negtive
            # we have e.g res = [A,4,A,4]
            elif len(res) == 4:
                chainID, resSeq = res[0], -int(res[1])

            if chainID not in resData.keys():
                resData[chainID] = []

            resData[chainID].append(resSeq)

        return resData

    ###################################################################
    #
    #   ROUTINES TO ACTUALY ALIGN THE MOLECULES
    #
    ###################################################################

    @staticmethod
    def compute_DockQScore(Fnat, lrmsd, irmsd, d1=8.5, d2=1.5):
        """Compute the DockQ Score

        Args:
            Fnat (float): Fnat value
            lrmsd (float): lrmsd value
            irmsd (float): irmsd value
            d1 (float, optional): first coefficient for the DockQ calculations
            d2 (float, optional): second coefficient for the DockQ calculations
        """
        def scale_rms(rms, d):
            return(1./(1+(rms/d)**2))

        return 1./3 * (Fnat + scale_rms(lrmsd, d1) + scale_rms(irmsd, d2))

    @staticmethod
    def get_rmsd(P, Q):
        """compute the RMSD

        Args:
            P (np.array(nx3)): position of the points in the first molecule
            Q (np.array(nx3)): position of the points in the second molecule

        Returns:
            float: RMSD value
        """
        n = len(P)
        return np.sqrt(1./n*np.sum((P-Q)**2))

    @staticmethod
    def get_trans_vect(P):
        """Get the translationv vector to the origin

        Args:
            P (np.array(nx3)): position of the points in the molecule

        Returns:
            float: minus mean value of the xyz columns
        """
        return -np.mean(P, 0)

    # main switch for the rotation matrix
    # add new methods here if necessary
    def get_rotation_matrix(self, P, Q, method='svd'):

        # get the matrix with Kabsh method
        if method.lower() == 'svd':
            return self.get_rotation_matrix_Kabsh(P, Q)

        # or with the quaternion method
        elif method.lower() == 'quaternion':
            return self.get_rotation_matrix_quaternion(P, Q)

        else:
            raise ValueError(
                '%s is not a valid method for rmsd alignement.\n Options are svd or quaternions' % method)

    @staticmethod
    def get_rotation_matrix_Kabsh(P, Q):
        '''Get the rotation matrix to aligh two point clouds.

        The method is based on th Kabsh approach
        https://cnx.org/contents/HV-RsdwL@23/Molecular-Distance-Measures

        Args:
            P (np.array): xyz of the first point cloud
            Q (np.array): xyz of the second point cloud

        Returns:
            np.array: rotation matrix

        Raises:
            ValueError: matrix have different sizes
        '''

        pshape = P.shape
        qshape = Q.shape

        if pshape[0] == qshape[0]:
            npts = pshape[0]
        else:
            raise ValueError(
                "Matrix don't have the same number of points", P.shape, Q.shape)

        p0, q0 = np.abs(np.mean(P, 0)), np.abs(np.mean(Q, 0))
        eps = 1E-6
        if any(p0 > eps) or any(q0 > eps):
            raise ValueError(
                'You must center the fragment first', p0, q0)

        # form the covariance matrix
        A = np.dot(P.T, Q)/npts

        # SVD the matrix
        V, S, W = np.linalg.svd(A)

        # the W matrix returned here is
        # already its transpose
        # https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.linalg.svd.html
        W = W.T

        # determinant
        d = np.linalg.det(np.dot(W, V.T))

        # form the U matrix
        Id = np.eye(3)
        if d < 0:
            Id[2, 2] = -1

        U = np.dot(W, np.dot(Id, V.T))

        return U

    @staticmethod
    def get_rotation_matrix_quaternion(P, Q):
        '''Get the rotation matrix to aligh two point clouds

        The method is based on the quaternion approach
        http://www.ams.stonybrook.edu/~coutsias/papers/rmsd17.pdf

        Args:
            P (np.array): xyz of the first point cloud
            Q (np.array): xyz of the second point cloud

        Returns:
            np.array: rotation matrix

        Raises:
            ValueError: matrix have different sizes
        '''

        pshape = P.shape
        qshape = Q.shape

        if pshape[0] != qshape[0]:
            raise ValueError(
                "Matrix don't have the same number of points", P.shape, Q.shape)

        p0, q0 = np.abs(np.mean(P, 0)), np.abs(np.mean(Q, 0))
        eps = 1E-6
        if any(p0 > eps) or any(q0 > eps):
            raise ValueError(
                'You must center the fragment first', p0, q0)

        # form the correlation matrix
        R = np.dot(P.T, Q)

        # form the F matrix (eq. 10 of ref[1])
        F = np.zeros((4, 4))

        F[0, 0] = np.trace(R)
        F[0, 1] = R[1, 2]-R[2, 1]
        F[0, 2] = R[2, 0]-R[0, 2]
        F[0, 3] = R[0, 1]-R[1, 0]

        F[1, 0] = R[1, 2]-R[2, 1]
        F[1, 1] = R[0, 0]-R[1, 1]-R[2, 2]
        F[1, 2] = R[0, 1]+R[1, 0]
        F[1, 3] = R[0, 2]+R[2, 0]

        F[2, 0] = R[2, 0]-R[0, 2]
        F[2, 1] = R[0, 1]+R[1, 0]
        F[2, 2] = -R[0, 0]+R[1, 1]-R[2, 2]
        F[2, 3] = R[1, 2]+R[2, 1]

        F[3, 0] = R[0, 1]-R[1, 0]
        F[3, 1] = R[0, 2]+R[2, 0]
        F[3, 2] = R[1, 2]+R[2, 1]
        F[3, 3] = -R[0, 0]-R[1, 1]+R[2, 2]

        # diagonalize it
        l, U = np.linalg.eig(F)

        # extract the eigenvect of the highest eigenvalues
        indmax = np.argmax(l)
        q0, q1, q2, q3 = U[:, indmax]

        # form the rotation matrix (eq. 33 ref[1])
        U = np.zeros((3, 3))

        U[0, 0] = q0**2+q1**2-q2**2-q3**2
        U[0, 1] = 2*(q1*q2-q0*q3)
        U[0, 2] = 2*(q1*q3+q0*q2)
        U[1, 1] = 2*(q1*q2+q0*q3)
        U[1, 2] = q0**2-q1**2+q2*2-q3**2
        U[1, 2] = 2*(q2*q3-q0*q1)
        U[2, 0] = 2*(q1*q3-q0*q2)
        U[2, 1] = 2*(q2*q3+q0*q1)
        U[2, 2] = q0**2-q1**2-q2**2+q3**2

        return U

    @staticmethod
    def translation(xyz, vect):
        """Translate a fragment

        Args:
            xyz (np.array): position of the fragment
            vect (np.array): translation vector

        Returns:
            np.array: translated positions
        """
        return xyz + vect

    @staticmethod
    def rotation_around_axis(xyz, axis, angle):
        """Rotate a fragment around an axis.

        Args:
            xyz (np.array): original positions
            axis (np.array): axis of rotation
            angle (float): angle of rotation (radians)

        Returns:
            np.array: Rotated positions
        """

        # get the data
        ct, st = np.cos(angle), np.sin(angle)
        ux, uy, uz = axis

        # get the center of the molecule
        xyz0 = np.mean(xyz, 0)

        # definition of the rotation matrix
        # see https://en.wikipedia.org/wiki/Rotation_matrix
        rot_mat = np.array([
            [ct + ux**2*(1-ct),         ux*uy*(1-ct) -
             uz*st,       ux*uz*(1-ct) + uy*st],
            [uy*ux*(1-ct) + uz*st,      ct + uy**2 *
             (1-ct),          uy*uz*(1-ct) - ux*st],
            [uz*ux*(1-ct) - uy*st,      uz*uy*(1-ct) + ux*st,       ct + uz**2*(1-ct)]])

        # apply the rotation
        return np.dot(rot_mat, (xyz-xyz0).T).T + xyz0

    @staticmethod
    def rotation_euler(xyz, alpha, beta, gamma):
        """Rotate a fragment from Euler rotation angle

        Args:
            xyz (np.array): original positions
            alpha (float): rotation angle around the x axis
            beta (float): rotation angle around the x axis
            gamma (float): rotation angle around the x axis

        Returns:
            np.array: Rotated positions
        """
        # precompute the trig
        ca, sa = np.cos(alpha), np.sin(alpha)
        cb, sb = np.cos(beta), np.sin(beta)
        cg, sg = np.cos(gamma), np.sin(gamma)

        # get the center of the molecule
        xyz0 = np.mean(xyz, 0)

        # rotation matrices
        rx = np.array([[1, 0, 0], [0, ca, -sa], [0, sa, ca]])
        ry = np.array([[cb, 0, sb], [0, 1, 0], [-sb, 0, cb]])
        rz = np.array([[cg, -sg, 0], [sg, cs, 0], [0, 0, 1]])

        rot_mat = np.dot(rx, np.dot(ry, rz))

        # apply the rotation
        return np.dot(rot_mat, (xyz-xyz0).T).T + xyz0

    @staticmethod
    def rotation_matrix(xyz, rot_mat, center=True):
        """Rotate a fragment from a roation matrix

        Args:
            xyz (np.array): original positions
            rot_mat (np.array): rotation matrix
            center (bool, optional): Center the fragment before rotation

        Returns:
            np.array: rotated positions
        """
        if center:
            xyz0 = np.mean(xyz)
            return np.dot(rot_mat, (xyz-xyz0).T).T + xyz0
        else:
            return np.dot(rot_mat, (xyz).T).T


# if __name__ == '__main__':
#     import time
#     BM4 = '/home/nico/Documents/projects/deeprank/data/HADDOCK/BM4_dimers/'
#     decoy = BM4 + 'decoys_pdbFLs/1AK4/water/1AK4_1w.pdb'
#     ref = BM4 + 'BM4_dimers_bound/pdbFLs_ori/1AK4.pdb'

#     sim = StructureSimilarity(decoy,ref)

#     #----------------------------------------------------------------------

#     t0 = time.time()
#     irmsd_fast = sim.compute_irmsd_fast(method='svd',izone='1AK4.izone')
#     t1 = time.time()-t0
#     print('\nIRMSD TIME FAST %f in %f sec' %(irmsd_fast,t1))

#     t0 = time.time()
#     irmsd = sim.compute_irmsd_pdb2sql(method='svd',izone='1AK4.izone')
#     t1 = time.time()-t0
#     print('IRMSD TIME SQL %f in %f sec' %(irmsd,t1))

#     #----------------------------------------------------------------------

#     t0 = time.time()
#     lrmsd_fast = sim.compute_lrmsd_fast(method='svd',lzone='1AK4.lzone',check=True)
#     t1 = time.time()-t0
#     print('\nLRMSD TIME FAST %f in %f sec' %(lrmsd_fast,t1))

#     t0 = time.time()
#     lrmsd = sim.compute_lrmsd_pdb2sql(exportpath=None,method='svd')
#     t1 = time.time()-t0
#     print('LRMSD TIME SQL %f in %f sec' %(lrmsd,t1))

#     #----------------------------------------------------------------------

#     t0 = time.time()
#     Fnat = sim.compute_Fnat_pdb2sql()
#     t1 = time.time()-t0
#     print('\nFNAT TIME SQL %f in %f sec' %(Fnat,t1))


#     t0 = time.time()
#     Fnat_fast = sim.compute_Fnat_fast(ref_pairs='1AK4.ref_pairs')
#     t1 = time.time()-t0
#     print('LRMSD TIME FAST %f in %f sec' %(Fnat_fast,t1))

#     #----------------------------------------------------------------------

#     dockQ = sim.compute_DockQScore(Fnat_fast,lrmsd_fast,irmsd_fast)
#     print('\nDockQ  %f' %dockQ )



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 20 12:27:54 2012


Taken from PAIRPred http://combi.cs.colostate.edu/supplements/pairpred/

@author: root
"""
"""
pdb_id: 
analysis type: unbound
radii_file: /home/fayyaz/PSAIA-1.0/amac_data/chothia.radii
ASA_zSice: 0.25
ASA_rSolvent: 1.4
rasa_file: /home/fayyaz/PSAIA-1.0/amac_data/natural_asa.asa
CX_sRadius: 10
CX_vAtom: 20.1
radii_file: /home/fayyaz/PSAIA-1.0/amac_data/hydrophobicity.hpb
total ASA: 0
back-bone ASA: 0
side-chain ASA: 0
polar ASA: 0
non-polar ASA: 0

  chain id|   ch total ASA|  ch b-bone ASA| ch s-chain ASA|   ch polar ASA|
   0             1               2               3               4 
  ch n-polar ASA|    res id|  res name|      total ASA|     b-bone ASA|   
  5                   6       7               8               9   
  s-chain ASA|      polar ASA|    n-polar ASA|     total RASA|    b-bone RASA|
  10              11              12              13              14          
   s-chain RASA|     polar RASA|   n-polar RASA|    average DPX|      s_avg DPX| 
   15                  16          17              18                  19      
   s-ch avg DPX| s-ch s_avg DPX|        max DPX|        min DPX|     average CX| 
     20          21                  22                  23          24  
   s_avg CX|    s-ch avg CX|  s-ch s_avg CX|  max CX|    min CX| Hydrophobicity|                                        
   25              26          27               28           29          30
   
"""
import tempfile
import os
import glob
import pdb

PSAIA_PATH='/home/nico/programs/PSAIA-1.0'

if PSAIA_PATH not in  os.environ["PATH"]:
    print('Adding '+PSAIA_PATH+' to system path')
    os.environ["PATH"]+=os.pathsep+PSAIA_PATH

def getFileParts(fname):
    "Returns the parts of a file"
    (path, name) = os.path.split(fname)
    n=os.path.splitext(name)[0]
    ext=os.path.splitext(name)[1]
    return (path,n,ext)
        
def createPSAIAConfig(opath=''):
    """
    Creates a dictionary object for the configuration file for PSAIA
    opath is the path to the output directory
    """
    psaia_dict={}
    psaia_dict['analyze_bound']='1'
    psaia_dict['analyze_unbound']='1'
    psaia_dict['calc_asa']='1'    
    psaia_dict['z_slice']='0.25'
    psaia_dict['r_solvent']='1.4'
    psaia_dict['write_asa']='1' 
    psaia_dict['calc_rasa']='1'    
    psaia_dict['standard_asa']=PSAIA_PATH+'/amac_data/natural_asa.asa'
    psaia_dict['calc_dpx']='1'
    psaia_dict['calc_cx']='1'
    psaia_dict['cx_threshold']='10'
    psaia_dict['cx_volume']='20.1'
    psaia_dict['calc_hydro']='1'
    psaia_dict['hydro_file']=PSAIA_PATH+'/amac_data/hydrophobicity.hpb'
    psaia_dict['radii_filename']=PSAIA_PATH+'/amac_data/chothia.radii'
    psaia_dict['write_xml']='0'
    psaia_dict['write_table']='1'
    psaia_dict['output_dir']=opath
    return psaia_dict
    
"""    
analyze_bound:	1
analyze_unbound:	1
calc_asa:	1
z_slice:	0.25
r_solvent:	1.4
write_asa:	1
calc_rasa:	1
standard_asa:	/s/chopin/c/proj/protfun/arch/x86_64/psaia/PSAIA-1.0/amac_data/natural_asa.asa
calc_dpx:	1
calc_cx:	1
cx_threshold:	10
cx_volume:	20.1
calc_hydro:	1
hydro_file:	/s/chopin/c/proj/protfun/arch/x86_64/psaia/PSAIA-1.0/amac_data/hydrophobicity.hpb
radii_filename:	/s/chopin/c/proj/protfun/arch/x86_64/psaia/PSAIA-1.0/amac_data/chothia.radii
write_xml:	1
write_table:	1
output_dir:	/s/chopin/c/proj/protfun/arch/x86_64/psaia/PSAIA-1.0
"""

def writePSAIAConfgFile(cfname,d):
    """
    Writes the configuration dictionary created in createPSAIAConfig to file cfname
    """
    f=open(cfname,'w')
    for k in d:
        s=k+':\t'+d[k]+'\n'
        f.write(s)
    f.close()
    

def runPSAIA(fname,ofname=None):
    """
    Given a pdb file (fname), this function returns all the psaia features in 
    a dictionary. It removes all the temporary files created when the ofname is None.
    If the output filename is specified the files are retained
    """
    rmv=False
    if ofname is None:
        ofname=fname
        rmv=True
    out_file = tempfile.NamedTemporaryFile(suffix='.psaia');  out_file.close()        
    flist=out_file.name
    out_file = tempfile.NamedTemporaryFile(suffix='.psaia');  out_file.close() 
    fconfig=out_file.name
    #print flist,fconfig
    #Make list file
    fh=open(flist,'w')
    fh.write(fname)
    fh.close()
    #Make config file
    (cdir,xf,_)=getFileParts(fname)
    if cdir=='':
        cdir=os.getcwd()
    writePSAIAConfgFile(fconfig,createPSAIAConfig(cdir))
    cmdstr='yes y | psa '+fconfig+' '+flist    
    #pdb.set_trace()
    ecode=os.system(cmdstr)    
    os.remove(flist)
    os.remove(fconfig)
    xfname=glob.glob(os.path.join(cdir,xf+'*bound.tbl'))#[0]
    print(xfname)
    pdict=make_psaia_dict(xfname)
    
    if rmv: #cleanup
        for x in glob.glob(os.path.join(cdir,xf+'*bound.tbl')):
            os.remove(x)
    print('PSAIA Successful : ', cmdstr,ecode,xfname)
    return pdict
    #else:
    #    print('PSAIA Processing Failed.')
    #    return 0
    
    
def make_psaia_dict(filename):

    """
    Return a stride dictionary that maps (chainid, resname, resid) to
    aa, ss and accessibility, from a stride output file.
    @param filename: the stride output file
    @type filename: string
    """

    psaia = {}
    stnxt=0
    ln=0
    try:        
        for l in open(filename, "r"):
            ln=ln+1
            ls=l.split()
            if stnxt:     
                cid=ls[0]
                if cid=='*': #psaia replaces cid ' ' in pdb files with *
                    cid=' '
                resid=(cid,ls[6])#cid,resid,resname is ignored
                casa=map(float,ls[1:6])
                rasa=map(float,ls[8:13])
                rrasa=map(float,ls[13:18])
                rdpx=map(float,ls[18:24])
                rcx=map(float,ls[24:30])
                rhph=float(ls[-1])
                psaia[resid]=(casa,rasa,rrasa,rdpx,rcx,rhph)
                #pdb.set_trace()
            elif len(ls) and ls[0]=='chain': # the line containing 'chain' is the last line before real data starts
                stnxt=1
            
    except Exception as e:
        print('Error Processing psaia file: ' ,filename)
        print('Error occured while processing line:',ln)
        print(e)
        raise(e)
    return psaia

if __name__=="__main__":
    psaia=runPSAIA('./1AK4.pdb')
    #psaia=make_psaia_dict()


import os
import numpy as np


def read_PSSM_data(fname):
    """Read the PSSM data."""

    f = open(fname,'r')
    data = f.readlines()
    f.close()

    filters = (lambda x: len(x.split())>0, lambda x: x.split()[0].isdigit())
    return list(map(lambda x: x.split(),list(filter(lambda x: all(f(x) for f in filters), data))))


def PSSM_aligned(pssmfiles,style='HADDOCK'):

    resmap = {
    'A' : 'ALA', 'R' : 'ARG', 'N' : 'ASN', 'D' : 'ASP', 'C' : 'CYS', 'E' : 'GLU', 'Q' : 'GLN',
    'G' : 'GLY', 'H' : 'HIS', 'I' : 'ILE', 'L' : 'LEU', 'K' : 'LYS', 'M' : 'MET', 'F' : 'PHE',
    'P' : 'PRO', 'S' : 'SER', 'T' : 'THR', 'W' : 'TRP', 'Y' : 'TYR', 'V' : 'VAL',
    'B' : 'ASX', 'U' : 'SEC', 'Z' : 'GLX'
    }

    pssm, ic = {}, {}
    for chain in ['A','B']:

        data = read_PSSM_data(pssmfiles[chain])
        for l in data :
            if style == 'res':
                resi = int(l[0])
                resn = resmap[l[1]]
            elif style == 'seq':
                resi = int(l[2])
                resn = resmap[l[3]]
            pssm[(chain,resi,resn)] = list(map(lambda x: float(x), l[4:24]))
            ic[(chain,resi,resn)] = float(l[24])

    return pssm, ic

def get_pssm_data(node,pssm):
    return pssm[node] if node in pssm else [0]*20

def get_ic_data(node,ic):
    return ic[node] if node in ic else 0.



import os
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.ResidueDepth import ResidueDepth, get_surface, residue_depth
from Bio.PDB.HSExposure import HSExposureCA

import warnings
from Bio import BiopythonWarning

import tempfile

with warnings.catch_warnings():
    warnings.simplefilter('ignore',BiopythonWarning)
    from Bio import SearchIO

from time import time

def get_bio_model(pdbfile):

    # (f,name) = tempfile.mkstemp()
    # sqldb.exportpdb(name)
    #parser = PDBParser(PERMISSIVE=1)
    parser = PDBParser()
    structure = parser.get_structure('_tmp',pdbfile)
    #os.remove(name)
    return structure[0]

def get_depth_res(model):

    t0 = time()
    rd = ResidueDepth(model)
    print('__ Create RD %f' %(time()-t0))

    data = {}
    t0 = time()
    for k in list(rd.keys()):
        new_key = (k[0],k[1][1])
        data[new_key] = rd[k][0]
    print('__ Reformat RD %f' %(time()-t0))
    return data


def get_depth_contact_res(model,contact_res):

    surface = get_surface(model)
    data = {}
    for r in contact_res:
        chain = model[r[0]]
        res = chain[r[1]]
        data[r] = residue_depth(res,surface)
    return data

def get_hse(model):
    hse = HSExposureCA(model)
    data = {}
    for k in list(hse.keys()):
        new_key = (k[0],k[1][1])

        x = hse[k]
        if x[2] is None:
            x = list(x)
            x[2] = 0.0
            x = tuple(x)

        data[new_key] = x
    return data



import unittest

from deeprank_gnn.NeuralNet import NeuralNet
from deeprank_gnn.ginet import GINet
from deeprank_gnn.foutnet import FoutNet


__PLOT__ = False


class TestNeuralNet(unittest.TestCase):

    def setUp(self):
        self.database = 'tests/hdf5/1ATN_residue.hdf5'

    def test_neural_net(self):
        NN = NeuralNet(self.database, GINet,
                       node_feature=['type', 'polarity', 'bsa',
                                     'depth', 'hse', 'ic', 'pssm'],
                       edge_feature=['dist'],
                       target='irmsd',
                       index=None,
                       task='reg',
                       batch_size=64,
                       percent=[0.8, 0.2])

        NN.train(nepoch=5, validate=False)

        NN.save_model('test.pth.tar')

        NN_cpy = NeuralNet(self.database, GINet,
                           pretrained_model='test.pth.tar')

        if __PLOT__:
            NN.plot_scatter()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import os

from setuptools import (find_packages, setup)

here = os.path.abspath(os.path.dirname(__file__))

# To update the package version number, edit DeepRank-GNN/__version__.py
version = {}
with open(os.path.join(here, 'deeprank_gnn', '__version__.py')) as f:
    exec(f.read(), version)

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(
    name='DeepRank-GNN',
    version=version['__version__'],
    description='Graph Neural network Scoring of protein-protein conformations',
    long_description=readme + '\n\n',
    long_description_content_type='text/markdown',
    author=["Nicolas Renaud", "Manon Reau"],
    author_email='n.renaud@esciencecenter.nl',
    url='https://github.com/DeepRank/DeepRank-GNN',
    packages=find_packages(),
    package_dir={'deeprank_gnn': 'deeprank_gnn'},
    include_package_data=True,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='deeprank_gnn',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English', 'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Chemistry'
    ],
    test_suite='tests',

    # not sure if the install of torhc-geometric will work ..
    install_requires=[
        'numpy >= 1.13', 'scipy', 'h5py', 'torch>=1.5.0', 'networkx',
        'pdb2sql', 'sklearn', 'chart-studio', 'BioPython', 'python-louvain',
        'markov-clustering', 'torch-sparse', 'torch-scatter', 'torch-cluster',
        'torch-spline-conv', 'torch-geometric'
    ],
    extras_require={
        'dev': ['prospector[with_pyroma]', 'yapf', 'isort'],
        'doc': ['recommonmark', 'sphinx', 'sphinx_rtd_theme'],
        'test':
        ['coverage', 'pycodestyle', 'pytest',
            'pytest-cov', 'pytest-runner', 'coveralls'],
    })



import os
import numpy as np
import pdb2sql

try:
    import freesasa

except ImportError:
    print('Freesasa not found')


class BSA(object):

    def __init__(self, pdb_data, sqldb, chainA='A', chainB='B'):
        '''Compute the burried surface area feature

        Freesasa is required for this feature.

        https://freesasa.github.io

        >>> wget http://freesasa.github.io/freesasa-2.0.2.tar.gz
        >>> tar -xvf freesasa-2.0.3.tar.gz
        >>> cd freesasa
        >>> ./configure CFLAGS=-fPIC (--prefix /home/<user>/)
        >>> make
        >>> make install

        Since release 2.0.3 the python bindings are separate module
        >>> pip install freesasa

        Args :
            pdb_data (list(byte) or str): pdb data or filename of the pdb
            chainA (str, optional): name of the first chain
            chainB (str, optional): name of the second chain

        Example :

        >>> bsa = BSA('1AK4.pdb')
        >>> bsa.get_structure()
        >>> bsa.get_contact_residue_sasa()
        >>> bsa.sql.close()

        '''

        self.pdb_data = pdb_data
        self.sql = sqldb
        self.chains_label = [chainA, chainB]

        freesasa.setVerbosity(freesasa.nowarnings)

    def get_structure(self):
        """Get the pdb structure of the molecule."""

        # we can have a str or a list of bytes as input
        if isinstance(self.pdb_data, str):
            self.complex = freesasa.Structure(self.pdb_data)
        else:
            self.complex = freesasa.Structure()
            atomdata = self.sql.get(
                'name,resName,resSeq,chainID,x,y,z')
            for atomName, residueName, residueNumber, chainLabel, x, y, z in atomdata:
                atomName = '{:>2}'.format(atomName[0])
                self.complex.addAtom(
                    atomName, residueName, residueNumber, chainLabel, x, y, z)
        self.result_complex = freesasa.calc(self.complex)

        self.chains = {}
        self.result_chains = {}
        for label in self.chains_label:
            self.chains[label] = freesasa.Structure()
            atomdata = self.sql.get(
                'name,resName,resSeq,chainID,x,y,z', chainID=label)
            for atomName, residueName, residueNumber, chainLabel, x, y, z in atomdata:
                atomName = '{:>2}'.format(atomName[0])
                self.chains[label].addAtom(
                    atomName, residueName, residueNumber, chainLabel, x, y, z)
            self.result_chains[label] = freesasa.calc(
                self.chains[label])

    def get_contact_residue_sasa(self, cutoff=8.5):
        """Compute the feature value."""

        self.bsa_data = {}
        self.bsa_data_xyz = {}

        res = self.sql.get_contact_residues(cutoff=cutoff)
        keys = list(res.keys())
        res = res[keys[0]]+res[keys[1]]

        for r in res:

            # define the selection string and the bsa for the complex
            select_str = ('res, (resi %d) and (chain %s)' %
                          (r[1], r[0]),)
            asa_complex = freesasa.selectArea(
                select_str, self.complex, self.result_complex)['res']

            # define the selection string and the bsa for the isolated
            select_str = ('res, resi %d' % r[1],)
            asa_unbound = freesasa.selectArea(
                select_str, self.chains[r[0]], self.result_chains[r[0]])['res']

            # define the bsa
            bsa = asa_unbound-asa_complex

            # define the xyz key : (chain,x,y,z)
            chain = {'A': 0, 'B': 1}[r[0]]
            xyz = np.mean(self.sql.get(
                'x,y,z', resSeq=r[1], chainID=r[0]), 0)
            xyzkey = tuple([chain]+xyz.tolist())

            # put the data in dict
            self.bsa_data[r] = [bsa]

#####################################################################################
#
#       TEST THE CLASS
#
#####################################################################################


if __name__ == '__main__':

    bsa = BSA('1AK4.pdb')
    bsa.get_structure()
    # bsa.get_contact_residue_sasa()
    bsa.sql.close()



import numpy as np
from time import time
import h5py
import os

# torch import
import torch
import torch.nn as nn
from torch.nn import MSELoss
import torch.nn.functional as F
from torch_geometric.data import DataLoader

# deeprank_gnn import
from .DataSet import HDF5DataSet, DivideDataSet, PreCluster
from .Metrics import Metrics


class NeuralNet(object):

    def __init__(self, database, Net,
                 node_feature=['type', 'polarity', 'bsa'],
                 edge_feature=['dist'], target='irmsd', lr=0.01,
                 batch_size=32, percent=[1.0, 0.0],
                 database_eval=None, index=None, class_weights=None, task=None,
                 classes=[0, 1], threshold=None,
                 pretrained_model=None, shuffle=True, outdir='./', cluster_nodes='mcl', transform_sigmoid=False):
        """Class from which the network is trained, evaluated and tested

        Args:
            database (str, required): path(s) to hdf5 dataset(s). Unique hdf5 file or list of hdf5 files.
            Net (function, required): neural network function (ex. GINet, Foutnet etc.)
            node_feature (list, optional): type, charge, polarity, bsa (buried surface area), pssm,
                    cons (pssm conservation information), ic (pssm information content), depth,
                    hse (half sphere exposure).
                    Defaults to ['type', 'polarity', 'bsa'].
            edge_feature (list, optional): dist (distance). Defaults to ['dist'].
            target (str, optional): irmsd, lrmsd, fnat, capri_class, bin_class, dockQ.
                    Defaults to 'irmsd'.
            lr (float, optional): learning rate. Defaults to 0.01.
            batch_size (int, optional): defaults to 32.
            percent (list, optional): divides the input dataset into a training and an evaluation set.
                    Defaults to [1.0, 0.0].
            database_eval ([type], optional): independent evaluation set. Defaults to None.
            index ([type], optional): index of the molecules to consider. Defaults to None.
            class_weights ([list or bool], optional): weights provided to the cross entropy loss function.
                    The user can either input a list of weights or let DeepRanl-GNN (True) define weights
                    based on the dataset content. Defaults to None.
            task (str, optional): 'reg' for regression or 'class' for classification . Defaults to None.
            classes (list, optional): define the dataset target classes in classification mode. Defaults to [0, 1].
            threshold (int, optional): threshold to compute binary classification metrics. Defaults to 4.0.
            pretrained_model (str, optional): path to pre-trained model. Defaults to None.
            shuffle (bool, optional): shuffle the training set. Defaults to True.
            outdir (str, optional): output directory. Defaults to ./
            cluster_nodes (bool, optional): perform node clustering ('mcl' or 'louvain' algorithm). Default to 'mcl'.
        """
        # load the input data or a pretrained model
        # each named arguments is stored in a member vairable
        # i.e. self.node_feature = node_feature
        if pretrained_model is None:
            for k, v in dict(locals()).items():
                if k not in ['self', 'database', 'Net', 'database_eval']:
                    self.__setattr__(k, v)

            if self.task == None:
                if self.target in ['irmsd', 'lrmsd', 'fnat', 'dockQ']:
                    self.task = 'reg'
                elif self.target in ['bin_class', 'capri_classes']:
                    self.task = 'class'
                else:
                    raise ValueError(
                        f"User target detected -> The task argument is required ('class' or 'reg'). \n\t"
                        f"Example: \n\t"
                        f""
                        f"model = NeuralNet(database, GINet,"
                        f"                  target='physiological_assembly',"
                        f"                  task='class',"
                        f"                  shuffle=True,"
                        f"                  percent=[0.8, 0.2])")
           
            if self.task == 'class' and self.threshold == None:
                print('the threshold for accuracy computation is set to {}'.format(self.classes[1]))
                self.threshold = self.classes[1]
            if self.task == 'reg' and self.threshold == None:
                print('the threshold for accuracy computation is set to 0.3')
                self.threshold = 0.3
            self.load_model(database, Net, database_eval)

        else:
            self.load_params(pretrained_model)
            self.outdir = outdir
            self.load_pretrained_model(database, Net)

    def load_pretrained_model(self, database, Net):
        """
        Loads pretrained model

        Args:
            database (str): path to hdf5 file(s)
            Net (function): neural network
        """
        # Load the test set
        test_dataset = HDF5DataSet(root='./', database=database,
                                   node_feature=self.node_feature, edge_feature=self.edge_feature,
                                   target=self.target)
        self.test_loader = DataLoader(
            test_dataset)
        PreCluster(test_dataset, method=self.cluster_nodes)

        print('Test set loaded')
        self.put_model_to_device(test_dataset, Net)

        self.set_loss()

        # optimizer
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.lr)

        # load the model and the optimizer state if we have one
        self.optimizer.load_state_dict(self.opt_loaded_state_dict)
        self.model.load_state_dict(self.model_load_state_dict)

    def load_model(self, database, Net, database_eval):
        """
        Loads model

        Args:
            database (str): path to the hdf5 file(s) of the training set
            Net (function): neural network
            database_eval (str): path to the hdf5 file(s) of the evaluation set

        Raises:
            ValueError: Invalid node clustering method.
        """
        # dataset
        dataset = HDF5DataSet(root='./', database=database, index=self.index,
                              node_feature=self.node_feature, edge_feature=self.edge_feature,
                              target=self.target)

        if self.cluster_nodes != None:
            if self.cluster_nodes == 'mcl' or self.cluster_nodes == 'louvain':
                print("Loading clusters")
                PreCluster(dataset, method=self.cluster_nodes)
            else:
                raise ValueError(
                    f"Invalid node clustering method. \n\t"
                    f"Please set cluster_nodes to 'mcl', 'louvain' or None. Default to 'mcl' \n\t")

        # divide the dataset
        train_dataset, valid_dataset = DivideDataSet(
            dataset, percent=self.percent)

        # dataloader
        self.train_loader = DataLoader(
            train_dataset, batch_size=self.batch_size, shuffle=self.shuffle)
        print('Training set loaded')

        if self.percent[1] > 0.0:
            self.valid_loader = DataLoader(
                valid_dataset, batch_size=self.batch_size, shuffle=self.shuffle)
            print('Evaluation set loaded')

        # independent validation dataset
        if database_eval is not None:
            print('Loading independent evaluation dataset...')
            valid_dataset = HDF5DataSet(root='./', database=database_eval, index=self.index,
                                        node_feature=self.node_feature, edge_feature=self.edge_feature,
                                        target=self.target)

            if self.cluster_nodes == 'mcl' or self.cluster_nodes == 'louvain':
                print('Loading clusters for the evaluation set.')
                PreCluster(valid_dataset, method=self.cluster_nodes)

            self.valid_loader = DataLoader(
                valid_dataset, batch_size=self.batch_size, shuffle=self.shuffle)
            print('Independent validation set loaded !')

        else:
            print('No independent validation set loaded')

        self.put_model_to_device(dataset, Net)

        # optimizer
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.lr)

        self.set_loss()

        # init lists
        self.train_acc = []
        self.train_loss = []

        self.valid_acc = []
        self.valid_loss = []

    def put_model_to_device(self, dataset, Net):
        """
        Puts the model on the available device

        Args:
            dataset (str): path to the hdf5 file(s)
            Net (function): neural network

        Raises:
            ValueError: Incorrect output shape
        """
        # get the device
        self.device = torch.device(
            'cuda' if torch.cuda.is_available() else 'cpu')

        print ('device set to :', self.device)
        if self.device.type == 'cuda':
            print(torch.cuda.get_device_name(0))

        # regression mode
        if self.task == 'reg':
            self.model = Net(dataset.get(
                0).num_features).to(self.device)

        # classification mode
        elif self.task == 'class':
            self.classes_to_idx = {i: idx for idx,
                                   i in enumerate(self.classes)}
            self.idx_to_classes = {idx: i for idx,
                                   i in enumerate(self.classes)}
            self.output_shape = len(self.classes)
            try:
                self.model = Net(dataset.get(
                    0).num_features, self.output_shape).to(self.device)
            except:
                raise ValueError(
                    f"The loaded model does not accept output_shape = {self.output_shape} argument \n\t"
                    f"Check your input or adapt the model\n\t"
                    f"Example :\n\t"
                    f"def __init__(self, input_shape): --> def __init__(self, input_shape, output_shape) \n\t"
                    f"self.fc2 = torch.nn.Linear(64, 1) --> self.fc2 = torch.nn.Linear(64, output_shape) \n\t")

    def set_loss(self):
        """Sets the loss function (MSE loss for regression/ CrossEntropy loss for classification)."""
        if self.task == 'reg':
            self.loss = MSELoss()

        elif self.task == 'class':

            # assign weights to each class in case of unbalanced dataset
            self.weights = None
            if self.class_weights == True:
                targets_all = []
                for batch in self.train_loader:
                    targets_all.append(batch.y)

                targets_all = torch.cat(
                    targets_all).squeeze().tolist()
                self.weights = torch.tensor(
                    [targets_all.count(i) for i in self.classes], dtype=torch.float32)
                print('class occurences: {}'.format(self.weights))
                self.weights = 1.0 / self.weights
                self.weights = self.weights / self.weights.sum()
                print('class weights: {}'.format(self.weights))

            self.loss = nn.CrossEntropyLoss(
                weight=self.weights, reduction='mean')

    def train(self, nepoch=1, validate=False, save_model='last', hdf5='train_data.hdf5', save_epoch='intermediate', save_every=5):
        """
        Trains the model

        Args:
            nepoch (int, optional): number of epochs. Defaults to 1.
            validate (bool, optional): perform validation. Defaults to False.
            save_model (last, best, optional): save the model. Defaults to 'last'
            hdf5 (str, optional): hdf5 output file
            save_epoch (all, intermediate, optional)
            save_every (int, optional): save data every n epoch if save_epoch == 'intermediate'. Defaults to 5
        """
        # Output file name
        fname = self.update_name(hdf5, self.outdir)

        # Open output file for writting
        self.f5 = h5py.File(fname, 'w')

        # Number of epochs
        self.nepoch = nepoch

        # Loop over epochs
        self.data = {}
        for epoch in range(1, nepoch+1):

            # Train the model
            self.model.train()

            t0 = time()
            _out, _y, _loss, self.data['train'] = self._epoch(epoch)
            t = time() - t0
            self.train_loss.append(_loss)
            self.train_out = _out
            self.train_y = _y
            _acc = self.get_metrics('train', self.threshold).accuracy
            self.train_acc.append(_acc)

            # Print the loss and accuracy (training set)
            self.print_epoch_data(
                'train', epoch, _loss, _acc, t)

            # Validate the model
            if validate is True:

                t0 = time()
                _out, _y, _val_loss, self.data['eval'] = self.eval(
                    self.valid_loader)
                t = time() - t0

                self.valid_loss.append(_val_loss)
                self.valid_out = _out
                self.valid_y = _y
                _val_acc = self.get_metrics(
                    'eval', self.threshold).accuracy
                self.valid_acc.append(_val_acc)

                # Print loss and accuracy (validation set)
                self.print_epoch_data(
                    'valid', epoch, _val_loss, _val_acc, t)

                # save the best model (i.e. lowest loss value on validation data)
                if save_model == 'best':

                    if min(self.valid_loss) == _val_loss:
                        self.save_model(filename='t{}_y{}_b{}_e{}_lr{}_{}.pth.tar'.format(
                            self.task, self.target, str(self.batch_size), str(nepoch), str(self.lr), str(epoch)))

            else:
                # if no validation set, saves the best performing model on the traing set
                if save_model == 'best':
                    if min(self.train_loss) == _train_loss:
                        print(
                            'WARNING: The training set is used both for learning and model selection.')
                        print(
                            'this may lead to training set data overfitting.')
                        print(
                            'We advice you to use an external validation set.')
                        self.save_model(filename='t{}_y{}_b{}_e{}_lr{}_{}.pth.tar'.format(
                            self.task, self.target, str(self.batch_size), str(nepoch), str(self.lr), str(epoch)))

            # Save epoch data
            if (save_epoch == 'all') or (epoch == nepoch):
                self._export_epoch_hdf5(epoch, self.data)

            elif (save_epoch == 'intermediate') and (epoch % save_every == 0):
                self._export_epoch_hdf5(epoch, self.data)

        # Save the last model
        if save_model == 'last':
            self.save_model(filename='t{}_y{}_b{}_e{}_lr{}.pth.tar'.format(
                self.task, self.target, str(self.batch_size), str(nepoch), str(self.lr)))

        # Close output file
        self.f5.close()

    def test(self, database_test=None, threshold=4, hdf5='test_data.hdf5'):
        """
        Tests the model

        Args:
            database_test ([type], optional): test database
            threshold (int, optional): threshold use to tranform data into binary values. Defaults to 4.
            hdf5 (str, optional): output hdf5 file. Defaults to 'test_data.hdf5'.
        """
        # Output file name
        fname = self.update_name(hdf5, self.outdir)

        # Open output file for writting
        self.f5 = h5py.File(fname, 'w')

        # Loads the test dataset if provided
        if database_test is not None:
            # Load the test set
            test_dataset = HDF5DataSet(root='./', database=database_test,
                                       node_feature=self.node_feature, edge_feature=self.edge_feature,
                                       target=self.target)
            print('Test set loaded')
            PreCluster(test_dataset, method='mcl')

            self.test_loader = DataLoader(
                test_dataset)

        else:
            if self.load_pretrained_model == None:
                raise ValueError(
                    "You need to upload a test dataset \n\t"
                    "\n\t"
                    ">> model.test(test_dataset)\n\t"
                    "if a pretrained network is loaded, you can directly test the model on the loaded dataset :\n\t"
                    ">> model = NeuralNet(database_test, gnn, pretrained_model = model_saved, target=None)\n\t"
                    ">> model.test()\n\t")
        self.data = {}

        # Run test
        _out, _y, _test_loss, self.data['test'] = self.eval(
            self.test_loader)

        self.test_out = _out

        if len(_y) == 0:
            self.test_y = None
            self.test_acc = None
        else:
            self.test_y = _y
            _test_acc = self.get_metrics('test', threshold).accuracy
            self.test_acc = _test_acc

        self.test_loss = _test_loss
        self._export_epoch_hdf5(0, self.data)

        self.f5.close()

    def eval(self, loader):
        """
        Evaluates the model

        Args:
            loader (DataLoader): [description]

        Returns:
            (tuple):
        """
        self.model.eval()

        loss_func, loss_val = self.loss, 0
        out = []
        y = []
        data = {'outputs': [], 'targets': [], 'mol': []}

        for data_batch in loader:

            data_batch = data_batch.to(self.device)
            pred = self.model(data_batch)
            pred, data_batch.y = self.format_output(
                pred, data_batch.y)

            # Check if a target value was provided (i.e. benchmarck scenario)
            if data_batch.y is not None:
                y += data_batch.y.tolist()
                loss_val += loss_func(pred,
                                      data_batch.y).detach().item()

            # get the outputs for export
            if self.task == 'class':
                pred = np.argmax(pred.detach(), axis=1)
            else:
                pred = pred.detach().reshape(-1)

            out += pred.tolist()

            # get the data
            data['mol'] += data_batch['mol']

        # Save targets
        if self.task == 'class':
            data['targets'] += [self.idx_to_classes[x]
                                for x in y]
            data['outputs'] += [self.idx_to_classes[x]
                                for x in out]

        else:
            data['targets'] += y
            data['outputs'] += out

        return out, y, loss_val, data

    def _epoch(self, epoch):
        """
        Runs a single epoch

        Returns:
            tuple: prediction, ground truth, running loss
        """
        running_loss = 0
        out = []
        y = []
        data = {'outputs': [], 'targets': [], 'mol': []}

        for data_batch in self.train_loader:
            data_batch = data_batch.to(self.device)
            self.optimizer.zero_grad()
            pred = self.model(data_batch)
            pred, data_batch.y = self.format_output(
                pred, data_batch.y)

            loss = self.loss(pred, data_batch.y)
            running_loss += loss.detach().item()
            loss.backward()
            self.optimizer.step()

            try:
                y += data_batch.y.tolist()
            except ValueError:
                print(
                    "You must provide target values (y) for the training set")

            # get the outputs for export
            if self.task == 'class':
                pred = np.argmax(pred.detach(), axis=1)
            else:
                pred = pred.detach().reshape(-1)

            out += pred.tolist()

            # get the data
            data['mol'] += data_batch['mol']

        # save targets and predictions
        if self.task == 'class':
            data['targets'] += [self.idx_to_classes[x]
                                for x in y]
            data['outputs'] += [self.idx_to_classes[x]
                                for x in out]
        else:
            data['targets'] += y
            data['outputs'] += out

        return out, y, running_loss, data

    def get_metrics(self, data='eval', threshold=4.0, binary=True):
        """
        Computes the metrics needed

        Args:
            data (str, optional): 'eval', 'train' or 'test'. Defaults to 'eval'.
            threshold (float, optional): threshold use to tranform data into binary values. Defaults to 4.0.
            binary (bool, optional): Transform data into binary data. Defaults to True.
        """
        if self.task == 'class':
            threshold = self.classes_to_idx[threshold]

        if data == 'eval':
            if len(self.valid_out) == 0:
                print('No evaluation set has been provided')

            else:
                pred = self.valid_out
                y = self.valid_y

        elif data == 'train':
            if len(self.train_out) == 0:
                print('No training set has been provided')

            else:
                pred = self.train_out
                y = self.train_y

        elif data == 'test':
            if len(self.test_out) == 0:
                print('No test set has been provided')

            if self.test_y == None:
                print(
                    'You must provide ground truth target values to compute the metrics')

            else:
                pred = self.test_out
                y = self.test_y

        return Metrics(pred, y, self.target, threshold, binary)

    def compute_class_weights(self):

        targets_all = []
        for batch in self.train_loader:
            targets_all.append(batch.y)

        targets_all = torch.cat(targets_all).squeeze().tolist()
        weights = torch.tensor([targets_all.count(i)
                                for i in self.classes], dtype=torch.float32)
        print('class occurences: {}'.format(weights))
        weights = 1.0 / weights
        weights = weights / weights.sum()
        print('class weights: {}'.format(weights))
        return weights

    @staticmethod
    def print_epoch_data(stage, epoch, loss, acc, time):
        """
        Prints the data of each epoch

        Args:
            stage (str): tain or valid
            epoch (int): epoch number
            loss (float): loss during that epoch
            acc (float or None): accuracy
            time (float): timing of the epoch
        """
        if acc is None:
            acc_str = 'None'
        else:
            acc_str = '%1.4e' % acc

        print('Epoch [%04d] : %s loss %e | accuracy %s | time %1.2e sec.' % (epoch,
                                                                             stage, loss, acc_str, time))

    def format_output(self, pred, target=None):
        """Format the network output depending on the task (classification/regression)."""
        if self.task == 'class' :
            pred = F.softmax(pred, dim=1)

            if target is not None:
                target = torch.tensor(
                    [self.classes_to_idx[int(x)] for x in target])

        elif self.transform_sigmoid is True :
            pred = torch.sigmoid(pred.reshape(-1))

        else:
            pred = pred.reshape(-1)

        return pred, target

    @staticmethod
    def update_name(hdf5, outdir):
        """
        Checks if the file already exists, if so, update the name

        Args:
            hdf5 (str): hdf5 file
            outdir (str): output directory

        Returns:
            str: update hdf5 name
        """
        fname = os.path.join(outdir, hdf5)

        count = 0
        hdf5_name = hdf5.split('.')[0]

        # If file exists, change its name with a number
        while os.path.exists(fname):
            count += 1
            hdf5 = '{}_{:03d}.hdf5'.format(hdf5_name, count)
            fname = os.path.join(outdir, hdf5)

        return fname

    def plot_loss(self, name=''):
        """
        Plots the loss of the model as a function of the epoch

        Args:
            name (str, optional): name of the output file. Defaults to ''.
        """
        nepoch = self.nepoch
        train_loss = self.train_loss
        valid_loss = self.valid_loss

        import matplotlib.pyplot as plt

        if len(valid_loss) > 1:
            plt.plot(range(1, nepoch+1), valid_loss,
                     c='red', label='valid')

        if len(train_loss) > 1:
            plt.plot(range(1, nepoch+1), train_loss,
                     c='blue', label='train')
            plt.title("Loss/ epoch")
            plt.xlabel("Number of epoch")
            plt.ylabel("Total loss")
            plt.legend()
            plt.savefig('loss_epoch{}.png'.format(name))
            plt.close()

    def plot_acc(self, name=''):
        """
        Plots the accuracy of the model as a function of the epoch

        Args:
            name (str, optional): name of the output file. Defaults to ''.
        """
        nepoch = self.nepoch
        train_acc = self.train_acc
        valid_acc = self.valid_acc

        import matplotlib.pyplot as plt

        if len(valid_acc) > 1:
            plt.plot(range(1, nepoch+1), valid_acc,
                     c='red', label='valid')

        if len(train_acc) > 1:
            plt.plot(range(1, nepoch+1), train_acc,
                     c='blue', label='train')
            plt.title("Accuracy/ epoch")
            plt.xlabel("Number of epoch")
            plt.ylabel("Accuracy")
            plt.legend()
            plt.savefig('acc_epoch{}.png'.format(name))
            plt.close()

    def plot_hit_rate(self, data='eval', threshold=4, mode='percentage', name=''):
        """
        Plots the hitrate as a function of the models' rank

        Args:
            data (str, optional): which stage to consider train/eval/test. Defaults to 'eval'.
            threshold (int, optional): defines the value to split into a hit (1) or a non-hit (0). Defaults to 4.
            mode (str, optional): displays the hitrate as a number of hits ('count') or as a percentage ('percantage') . Defaults to 'percentage'.
        """
        import matplotlib.pyplot as plt

        try:

            hitrate = self.get_metrics(data, threshold).hitrate()

            nb_models = len(hitrate)
            X = range(1, nb_models + 1)

            if mode == 'percentage':
                hitrate /= hitrate.sum()

            plt.plot(X, hitrate, c='blue', label='train')
            plt.title("Hit rate")
            plt.xlabel("Number of models")
            plt.ylabel("Hit Rate")
            plt.legend()
            plt.savefig('hitrate{}.png'.format(name))
            plt.close()

        except:
            print('No hit rate plot could be generated for you {} task'.format(
                self.task))

    def plot_scatter(self):
        """Scatters plot of the results."""
        import matplotlib.pyplot as plt

        self.model.eval()

        pred, truth = {'train': [], 'valid': []}, {
            'train': [], 'valid': []}

        for data in self.train_loader:
            data = data.to(self.device)
            truth['train'] += data.y.tolist()
            pred['train'] += self.model(data).reshape(-1).tolist()

        for data in self.valid_loader:
            data = data.to(self.device)
            truth['valid'] += data.y.tolist()
            pred['valid'] += self.model(data).reshape(-1).tolist()

        plt.scatter(truth['train'], pred['train'], c='blue')
        plt.scatter(truth['valid'], pred['valid'], c='red')
        plt.show()

    def save_model(self, filename='model.pth.tar'):
        """
        Saves the model to a file

        Args:
            filename (str, optional): name of the file. Defaults to 'model.pth.tar'.
        """
        state = {'model': self.model.state_dict(),
                 'optimizer': self.optimizer.state_dict(),
                 'node': self.node_feature,
                 'edge': self.edge_feature,
                 'target': self.target,
                 'task': self.task,
                 'classes': self.classes,
                 'class_weight': self.class_weights,
                 'batch_size': self.batch_size,
                 'percent': self.percent,
                 'lr': self.lr,
                 'index': self.index,
                 'shuffle': self.shuffle,
                 'threshold': self.threshold,
                 'cluster_nodes': self.cluster_nodes,
                 'transform_sigmoid': self.transform_sigmoid}

        torch.save(state, filename)

    def load_params(self, filename):
        """
        Loads the parameters of a pretrained model

        Args:
            filename ([type]): [description]

        Returns:
            [type]: [description]
        """
        self.device = torch.device(
            'cuda' if torch.cuda.is_available() else 'cpu')

        state = torch.load(filename, map_location=torch.device(self.device))

        self.node_feature = state['node']
        self.edge_feature = state['edge']
        self.target = state['target']
        self.batch_size = state['batch_size']
        self.percent = state['percent']
        self.lr = state['lr']
        self.index = state['index']
        self.class_weights = state['class_weight']
        self.task = state['task']
        self.classes = state['classes']
        self.threshold = state['threshold']
        self.shuffle = state['shuffle']
        self.cluster_nodes = state['cluster_nodes']
        self.transform_sigmoid = state['transform_sigmoid']

        self.opt_loaded_state_dict = state['optimizer']
        self.model_load_state_dict = state['model']

    def _export_epoch_hdf5(self, epoch, data):
        """
        Exports the epoch data to the hdf5 file.

        Exports the data of a given epoch in train/valid/test group.
        In each group are stored the predicted values (outputs),
        ground truth (targets) and molecule name (mol).

        Args:
            epoch (int): index of the epoch
            data (dict): data of the epoch
        """
        # create a group
        grp_name = 'epoch_%04d' % epoch
        grp = self.f5.create_group(grp_name)

        grp.attrs['task'] = self.task
        grp.attrs['target'] = self.target
        grp.attrs['batch_size'] = self.batch_size

        # loop over the pass_type : train/valid/test
        for pass_type, pass_data in data.items():

            # we don't want to breack the process in case of issue
            try:

                # create subgroup for the pass
                sg = grp.create_group(pass_type)

                # loop over the data : target/output/molname
                for data_name, data_value in pass_data.items():

                    # mol name is a bit different
                    # since there are strings
                    if data_name == 'mol':
                        data_value = np.string_(data_value)
                        string_dt = h5py.special_dtype(vlen=str)
                        sg.create_dataset(
                            data_name, data=data_value, dtype=string_dt)

                    # output/target values
                    else:
                        sg.create_dataset(data_name, data=data_value)

            except TypeError:
                raise ValueError("Error in export epoch to hdf5")
